""" Modifed version of calnexRest.py
This file provides an interface to Paragon-neo and Paragon-100G.
It has been modified to add a number of functoions required to provide
support for Paragon-X style commands
"""
##############################################################################
# Copyright (c) Calnex Solutions Ltd 2008 - 2018
#
# The contents of this file should not be copied or distributed
# without permission being granted by Calnex Solutions Ltd.
#
# All rights reserved.
##############################################################################

import os
import time
import json
import requests

_helper = None


class CalnexRestHelper():
    """ Helper class
    """
    _instrument_type = ""
    _ip_address = ""
    instrument = ""
    last_err = ""

    def __init__(self, ip_address, inst):
        if inst == "SNE":
            self.instrument = "http://" + ip_address + ":5000/"
        else:
            self.instrument = "http://" + ip_address + "/api/"

        self._ip_address = ip_address
        self._instrument_type = inst

    def set_instrument_type(self, inst=""):
        if inst == "":
            self._instrument_type = calnexGetVal(
                "instrument/information", "HwType")
        else:
            self._instrument_type = inst

        if inst == "SNE":
            self.instrument = "http://" + self._ip_address + ":5000/"
        else:
            self.instrument = "http://" + self._ip_address + "/api/"

    def calnex_cat_get_missing_count(self, from_slot):
        """ Returns the number of missing pulses / packets in a capture
        Arguments:
            from_slot   The name of the slot from which the count should come.
                        One of "1pps", "2Way", "Sync", "Delay Req"
        Results:
            Returns the number of missing pulses / packets
        """
        retval = 0
        supported_slots = {"1pps", "2Way", "Sync", "Delay Req"}
        slot_to_rest_mapping = {
            "1pps": {"Metric": "1ppsTEAbsolute", "Slot": "ONEPPS"},
            "2Way": {"Metric": "TBD", "Slot": "TBD"},
            "Sync": {"Metric": "TBD", "Slot": "TBD"},
            "Delay Req": {"Metric": "TBD", "Slot": "TBD"}}

        if any(slot in from_slot for slot in supported_slots):
            metric = slot_to_rest_mapping[from_slot]["Metric"]
            slot = slot_to_rest_mapping[from_slot]["Slot"]
            url = \
                "/cat/measurement/" + metric + \
                "/-/" + slot + "/-/series/Errors"
            try:
                resp = calnexGetVal(url, "Plots")
                plots = len(resp)
                if plots > 0:
                    pts = resp[0]
                    points = pts["Points"]
                    retval = len(points)
            except Exception:
                # If the capture has no missing pulses, an exception is thrown.
                # We can't be sure there are none but it is a reasonable
                # assumption
                retval = 0
        else:
            raise ValueError("Invalid argument: " +
                             from_slot + " not supported")

        return retval

    def calnex_cat_done(self):
        """ Wait for the CAT to finish opening files and processing data
        Arguments:
            None
        Results:
            Returns True when the CAT no longer indicates that it is
            processing or opening a file.
            The polling period is 1 second with a meximum of 60 re-tries.
            If the re-try count is exceeded, False is returned.
        """
        cat_status = calnexGet("/cat/general/status")
        cat_currently_processing = cat_status["IsApiCurrentlyProcessing"]
        cat_opening = cat_status["IsOpeningInProgress"]
        retry = 0

        while cat_currently_processing or cat_opening:
            time.sleep(1)
            cat_status = calnexGet("/cat/general/status")
            cat_currently_processing = cat_status["IsApiCurrentlyProcessing"]
            cat_opening = cat_status["IsOpeningInProgress"]
            retry = retry + 1
            if retry > 60:
                return False
        return True

    def calnex_cat_generate_report(self, with_charts,
                                   report_name, dest_folder="./"):
        """ Generate a report in the CAT and then download it to the local PC
        Parameters:
            reportName: str
                The name of the report to be generated
            withCharts: bool
                If TRUE, then charts will be included in the report.
            destFolder: str, optional
                The name of the folder on the local PC where the report will
                be saved. The path to the folder will be created if required.
                If destFolder is not specified then the report will be
                saved in the current working directory (i.e. where
                the script is executing)
        Returns:
            None
        Raises:
            Raises a runtime exception if the CAT remains busy
        """
        if self.calnex_cat_done():
            calnexGet("/cat/report/data")
            calnexCreate("/cat/report", "RenderCharts", with_charts,
                         "ReportFilename", report_name)

            # Report is now generated. Download it.
            calnex_download_file("ReportFolder", "./", report_name, dest_folder)

        else:
            raise RuntimeError(
                "Unable to generate report. CAT is still processing.")

    def check_for_error(self, label):
        """
        Raise an exception is the last operation resulted in an error
        """
        if self.last_err != "":
            raise Exception("%s : %s" % (label, self.last_err))

    def calnex_cat_get_cte(self, from_slot):
        """
        Returns CTE from the specified slot
        """
        retval = 0
        supported_slots = {"1pps", "2Way", "Sync", "DelayReq"}
        slot_to_rest_mapping = {
            "1pps": {"Metric": "1ppsTEAbsolute", "Port": "F"},
            "2Way": {"Metric": "2Way", "Port": "D"},
            "Sync": {"Metric": "Sync", "Port": "D"},
            "DelayReq": {"Metric": "DelayReq", "Port": "D"}}

        if any(slot in from_slot for slot in supported_slots):
            metric = slot_to_rest_mapping[from_slot]["Metric"]
            if self._instrument_type == "Paragon-X":
                port = "-"
            else:
                port = slot_to_rest_mapping[from_slot]["Port"]

            # Check that the metric is enabled
            url = "/cat/measurement/" + metric + \
                "/" + port + "/" + "CTE" + "/-/isenabled"
            enabled = calnexGetVal(url, "Value")
            if enabled:
                url = "/cat/measurement/" + metric + \
                    "/" + port + "/" + "CTE" + "/-/statistics/cTe"
                resp = calnexGetVal(url, "StatisticsValue")
                retval = resp
            else:
                str1 = metric + " CTE is not enabled"
                raise RuntimeError(str1)
        else:
            if from_slot == "NA":
                str1 = "Invalid argument: no slot has been selected."
                str2 = "You must call Cat SelectSlot before this command."
                raise ValueError(str1 + " " + str2)
            else:
                raise ValueError(
                    "Invalid argument: " + from_slot + " not supported")

        return retval


def args_to_json(arg):
    """
    Convert from list to JSON
    """
    i = iter(arg)
    dic = dict(zip(i, i))
    return json.dumps(dic)


def px_init(ip_addr):
    """
    calnexRest is designed for Paragon-100G and Paragon-Neo but
    it can be used to access the CAT with Paragon-X. This function
    configures the wrapper to do this
    """
    global _helper

    if ip_addr == "":
        raise Exception("%s : %s" % (
            "pxInit", "Must specify an IP Address for the instrument"))
    else:
        _helper = CalnexRestHelper(ip_addr, "Paragon-X")

    return _helper


def sne_init(ip_addr):
    global _helper

    if ip_addr == "":
        raise Exception("%s : %s" % (
            "pxInit", "Must specify an IP Address for the instrument"))
    else:
        _helper = CalnexRestHelper(ip_addr, "SNE")

    return _helper


def calnexInit(ip_addr):
    """
    Initialise the wrapper
    """
    global _helper

    if ip_addr == "":
        raise Exception("%s : %s" % (
            "calnexInit", "Must specify an IP Address for the instrument"))

    else:
        _helper = CalnexRestHelper(ip_addr, "")
        _helper.set_instrument_type()
        _helper.check_for_error("calnexInit")

    return _helper


def calnexGet(url, *arg):
    """
    Read from the connected instrument
    """
    # global _instrument
    # global _last_err

    _helper.last_err = ""
    if _helper.instrument == "":
        _helper.last_err = \
            "IP address not configured - call calnexInit before anything else"
        ret = ""
    else:
        try:
            response = requests.get(
                "{0}{1}".format(_helper.instrument, url),
                data=args_to_json(arg),
                headers={'Content-Type': 'application/json'})
            response.raise_for_status()
            ret = response.json()
        except requests.exceptions.RequestException as exc:
            _helper.last_err = str(exc)

    _helper.check_for_error("calnexGet %s" % (url))
    return ret


def calnexGetVal(url, arg):
    """
    Read from the connected isntrument and return a specified setting
    """
    # global _instrument
    # global _last_err
    res = calnexGet(url, arg)
    ret = res
    if arg not in res:
        _helper.last_err = \
            "\"" + arg + "\" does not exist in response: " + str(res)
    else:
        ret = res[arg]

    _helper.check_for_error("calnexGetVal %s %s" % (url, arg))
    return ret


def calnexSet(url, *arg):
    """
    Write to a setting in the connected isntrument
    """
    # global _instrument
    # global _last_err
    ret = ""

    _helper.last_err = ""
    if _helper.instrument == "":
        _helper.last_err = \
            "IP address not configured - call calnexInit before anything else"
        ret = ""
    else:
        try:
            response = requests.put(
                "{0}{1}".format(_helper.instrument, url),
                args_to_json(arg),
                headers={'Content-Type': 'application/json'})
            response.raise_for_status()

            if response.text != "":
                ret = response.json()
            else:
                ret = ""
                # print (ret)
        except requests.exceptions.RequestException as exc:
            _helper.last_err = str(exc)
            _helper.check_for_error("calnexSet %s" % (url))
    return ret


def calnexCreate(url, *arg):
    """
    TBD
    """
    requests.post(
        "{0}{1}".format(_helper.instrument, url),
        args_to_json(arg),
        headers={'Content-Type': 'application/json'}).raise_for_status()


def calnexDel(url):
    """
    TBD
    """
    # global _instrument
    requests.delete(
        "{0}{1}".format(_helper.instrument, url),
        headers={'Content-Type': 'application/json'}).raise_for_status()

#
# Old syntax - kept for backwards compatibility
#


def p100get(url, *arg):
    """
    Retained for backwards compatibility only
    """
    return calnexGet(url, *arg)


def p100set(url, *arg):
    """
    Retained for backwards compatibility only
    """
    calnexSet(url, *arg)


def p100create(url, *arg):
    """
    Retained for backwards compatibility only
    """
    calnexCreate(url, *arg)


def p100del(url):
    """
    Retained for backwards compatibility only
    """
    calnexDel(url)


def a100get(url, *arg):
    """
    Retained for backwards compatibility only
    """
    return calnexGet(url, *arg)


def a100set(url, *arg):
    """
    Retained for backwards compatibility only
    """
    calnexSet(url, *arg)


def a100create(url, *arg):
    """
    Retained for backwards compatibility only
    """
    calnexCreate(url, *arg)


def a100del(url):
    """
    Retained for backwards compatibility only
    """
    calnexDel(url)


# Utility functions and associated globals ####################################

def calnex_upload_file(file_type, full_path):
    """ Upload a file to the currently connected instrument

    Parameters:
        file_type:
            Determines where the file is stored on the instrument and
            the upload command to use.
            The allowable file_types vary by instrument (see below)
        full_path:
            The path to the file on the local machine.
            Must include the filename and extension.
            The filename of the source file is used as the filename of the
            uploaded file.

    Notes:
        (1) For JitterProfile, any existing files with the same name will be
            overwritten

    Allowable file_types:
        Paragon: PfvUserRules
        Attero: JitterProfile
    """
    model = calnexGetVal("instrument/information", "HwType")

    if model == "Attero-100G":
        resp = _attero_upload_file(file_type, full_path)
    elif model == "Paragon-100G":
        resp = _paragon_upload_file(file_type, full_path)
    else:
        raise Exception(
            "%s %s %s" % ("calnexUploadFile: ", model, " not recognised"))

    return resp


def _attero_upload_file(file_type, full_path):
    """ Attero specific uploads """
    resp = None
    allowable_file_types = {"JitterProfile", "Pcap"}
    if file_type not in allowable_file_types:
        raise Exception(
            "%s %s %s" % ("calnexUploadFile: ", file_type, " not recognised"))

    if file_type == "JitterProfile":
        url = "/app/impairment/jitter/jitterprofile/file/upload?Overwrite=true"
        resp = _upload(url, full_path)
    # elif file_type == "Pcap":
    #    url = "/filter/pcap/analysis/upload"
    #    resp = _upload(url, full_path)

    return resp


def _paragon_upload_file(file_type, full_path):
    """ Paragon specific uploads """
    raise Exception("File upload to Paragon is not supported")

    resp = None
    allowable_file_types = {"PfvUserRules"}
    if file_type not in allowable_file_types:
        raise Exception(
            "%s %s %s" % ("calnexUploadFile: ", file_type, " not recognised"))

    name = os.path.basename(full_path)
    url = "cat/filecommander/PfvUserRulesFolder?FileId=" + name
    resp = _upload(url, full_path)
    # files = {'file': open(full_path, 'rb')}
    # r = requests.post("{0}{1}".format(_instrument, url), files=files)
    return resp


def sne_upload_file(full_path):
    """ TBD """
    resp = None
    # name = os.path.basename(full_path)
    url = "maps/upload"
    resp = _upload(url, full_path)
    return resp


def _upload(url, full_path):
    """
    """
    files = {'file': open(full_path, 'rb')}
    # Needs to be a put for SNE
    resp = requests.post(
        "{0}{1}".format(_helper.instrument, url),
        files=files).raise_for_status()
    print(resp)
    return resp


def calnex_attero_apply_jitter_profile(filename, port, position):
    """ calnexAtteroApplyJitterProfile

    Apply a jitter profile.
    The profile file must exist on the instrument.
    This is the equivalent of the "Extract" button on the UI.

    Parameters
    filename:
        The filename of the profile file. This must exist in the
        "Patterns/Jitter" folder on the instrument
    port:
        "Port1" or "Port2"
    position:
        "0" to "7"
    """
    url = "app/impairment/jitter/jitterprofile/file/extract?FileName="
    jitter_profile = calnexGet(url + filename)
    calnexCreate(
        "app/impairment/jitter/{0}/{1}/jitterprofile".format(port, position),
        "JitterProfile", jitter_profile)


def calnex_download_file(folder_type, src_folder, filename, dest_folder):
    """ Download a file from the instrument
    Arguments:
        folderType	"SessionsFolder" or "ReportFolder"
        srcFolder	The name of the folder on the instrument. For sessions files
                    this is the name of the session folder e.g. Session_<date>
        file 		The name of the file - for capture files, this is the name of
                    the file in the Session folder
        destFolder	The name of the folder on the local machines where the
                    remote folder/file will be saved
    Results:
        Raises an error if the file cannot be found on the isntrument
        If the local file or folder can't be accessed then tcl will raise a
        file access error
    """
    # raise Exception("Not currently supported")

    remote_file = os.path.join(src_folder, filename)
    url = \
        "cat/filecommander/download/" + folder_type + \
        "?AsAttachment=true&FileId=" + remote_file
    local_file = os.path.join(dest_folder, filename)
    local_fid = open(local_file, "wb")

    try:
        response = requests.get("{0}{1}".format(_helper.instrument, url))
        response.raise_for_status()
        local_fid.write(response.content)
    except requests.exceptions.RequestException as exc:
        _helper.last_err = str(exc)

    local_fid.close()

    _helper.check_for_error(
        "calnexDownloadFile: Unable to download " + filename
        + " from " + src_folder)

    return
